"""check the weather"""
from collections import defaultdict
from datetime import datetime, timedelta

from apscheduler.schedulers.asyncio import AsyncIOScheduler     # type: ignore

from wechaty import Message
from wechaty.plugin import WechatyPlugin


class DingDongRankPlugin(WechatyPlugin):
    """weather plugin for bot"""
    def __init__(self):
        """init the plugin"""
        super().__init__()

        self.room_data = {}
        self.interval_time = 8

        self.history_analyze = {}
        self.top_k = 10
        self.room_alias = {}

        self.scheduler = AsyncIOScheduler()
        self.scheduler.start()

    @property
    def name(self) -> str:
        """get the name of the plugin"""
        return 'ding-dong-rank'

    async def on_message(self, msg: Message):
        """listen message event"""
        from_contact = msg.talker()
        text = msg.text()
        room = msg.room()

        if room is None:
            return

        if text in ['#ding', 'dong']:
            if room.room_id not in self.room_data:
                self.room_data[room.room_id] = {
                    '__status__': 'running'
                }
                await self.re_init_job(room.room_id)
        if text == '#ding':
            self.room_data[room.room_id]['__ding__'] = True

        if text == '#ding start':
            self.room_data[room.room_id] = {
                '__status__': 'running'
            }
            await room.say('#ding')
            await self.re_init_job(room.room_id)

        if text == 'dong':
            if room.room_id in self.room_data:
                self.room_data[room.room_id][from_contact.contact_id] = {
                    'id': from_contact.contact_id,
                    'time': datetime.now(),
                    'name': from_contact.name,
                }
                # refresh the scheduler job
                await self.re_init_job(room.room_id)

    async def get_room_alias(self, room_id, member_id, name):
        """get alias in room"""
        room = self.bot.Room(room_id)
        await room.ready()
        members = await room.member_all()

        if room_id not in self.room_alias:
            self.room_alias[room_id] = {}
            for member in members:
                await member.ready()
                alias = await room.alias(member)
                print('---' + alias + '---')
                if alias is None or alias == '':
                    self.room_alias[room_id][member.contact_id] = \
                        member.payload.name
                else:
                    self.room_alias[room_id][member.contact_id] = alias

        if member_id not in self.room_alias[room_id]:
            return name
        return self.room_alias[room_id][member_id]

    async def send_rank_analysis(self, room_id):
        """send rank analysis data to conversation room"""
        if room_id not in self.room_data:
            return

        # check if contains ding info
        if not self.room_data[room_id].get('__ding__', False):
            return

        bot_ids = [key for key in self.room_data[room_id].keys() if
                   not key.startswith('__')]
        dong_data = [self.room_data[room_id][bot_id] for bot_id in bot_ids]

        # sort the date
        dong_data.sort(key=lambda x: x['time'])

        if room_id not in self.history_analyze:
            self.history_analyze[room_id] = {}

        top_k_result = []
        last_timestamp = datetime(year=2000, month=1, day=1)
        top = 0
        for index, item in enumerate(dong_data):
            if item['id'] not in self.history_analyze[room_id]:
                self.history_analyze[room_id][item['id']] = defaultdict()

            self.history_analyze[room_id][item['id']][index + 1] = \
                self.history_analyze[room_id][item['id']].get(index + 1, 0) + 1
            if index <= self.top_k:
                if item['time'] > last_timestamp:
                    last_timestamp = item['time']
                    top += 1
                name = await self.get_room_alias(
                    room_id, item['id'], item['name']
                )
                top_k_result.append({
                    'id': item['id'],
                    'name': name,
                    'top': top
                })
        # send message to the room
        msg = '🔥🔥 DING-DONG Speed Ranking \n=========================\n'

        icon = {
            1: '1️⃣', 2: '2️⃣', 3: '3️⃣', 4: '4️⃣', 5: '5️⃣',
            6: '6️⃣', 7: '7️⃣', 8: '8️⃣', 9: '9️⃣', 10: '🔟'
        }

        for item in top_k_result:
            if item['top'] in icon:
                msg += f'NO.{icon[item["top"]]} {item["name"]}\n'

        room = self.bot.Room.load(room_id)
        await room.ready()
        await room.say(msg)

        del self.room_data[room_id]

    async def re_init_job(self, room_id):
        """re init job, replace it if exist"""
        job_id = f'__job__{room_id}'
        now = datetime.now()

        job = self.scheduler.add_job(
            self.send_rank_analysis,
            replace_existing=True,
            id=job_id,
            trigger='date',
            next_run_time=now + timedelta(seconds=self.interval_time),
            kwargs={'room_id': room_id}
        )
        print('add job ...')
        print(job)



"""
Python Wechaty - https://github.com/wechaty/python-wechaty

Authors:    Huan LI (李卓桓) <https://github.com/huan>
            Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2020-now @ Copyright Wechaty

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations

from abc import abstractmethod, ABCMeta
from enum import Enum
from typing import (
    TYPE_CHECKING,
    List,
    Optional,
    Dict,
    Union,
    Any)
import re
from datetime import datetime
from copy import deepcopy
from dataclasses import dataclass
from collections import defaultdict, OrderedDict
from wechaty_puppet import get_logger   # type: ignore

if TYPE_CHECKING:
    from wechaty_puppet import (
        EventErrorPayload,
        EventHeartbeatPayload,
        EventReadyPayload,
        ScanStatus
    )
    from wechaty import (
        Room,
        Friendship,
        Contact,
        Message,
        Wechaty
    )
    from wechaty.user.room_invitation import RoomInvitation


log = get_logger(__name__)


@dataclass
class WechatyPluginOptions:
    """options for wechaty plugin"""
    name: Optional[str] = None
    metadata: Optional[dict] = None


class PluginStatus(Enum):
    """plugin running status"""
    Running = 0
    Stopped = 1


class WechatyPlugin(metaclass=ABCMeta):
    """
    abstract wechaty plugin base class

    listen events from

    """
    def __init__(self, options: Optional[WechatyPluginOptions] = None):
        self.output: Dict[str, Any] = {}
        self.bot: Optional[Wechaty] = None
        if options is None:
            options = WechatyPluginOptions()
        self.options = options

    async def init_plugin(self, wechaty: Wechaty):
        """set wechaty to the plugin"""
        self.bot = wechaty

    @property
    @abstractmethod
    def name(self) -> str:
        """you must give a name for wechaty plugin"""
        raise NotImplementedError

    @staticmethod
    def get_dependency_plugins() -> List[str]:
        """
        get dependency plugins
        """
        return []

    async def on_error(self, payload: EventErrorPayload):
        """
        listen error event for puppet

        this is friendly for code typing
        """

    async def on_heartbeat(self, payload: EventHeartbeatPayload):
        """
        listen heartbeat event for puppet

        this is friendly for code typing
        """

    async def on_friendship(self, friendship: Friendship):
        """
        listen friendship event for puppet

        this is friendly for code typing
        """

    async def on_login(self, contact: Contact):
        """
        listen login event for puppet

        this is friendly for code typing
        """

    async def on_logout(self, contact: Contact):
        """
        listen logout event for puppet

        this is friendly for code typing
        """

    async def on_message(self, msg: Message):
        """
        listen message event for puppet

        this is friendly for code typing
        """

    async def on_ready(self, payload: EventReadyPayload):
        """
        listen ready event for puppet

        this is friendly for code typing
        """

    async def on_room_invite(self, room_invitation: RoomInvitation):
        """
        listen room_invitation event for puppet

        this is friendly for code typing
        """

    async def on_room_join(self, room: Room, invitees: List[Contact],
                           inviter: Contact, date: datetime):
        """
        listen room_join event for puppet

        this is friendly for code typing
        """

    async def on_room_leave(self, room: Room, leavers: List[Contact],
                            remover: Contact, date: datetime):
        """
        listen room_leave event for puppet

        room, leavers, remover, date

        this is friendly for code typing
        """

    # pylint: disable=R0913
    async def on_room_topic(self, room: Room, new_topic: str, old_topic: str,
                            changer: Contact, date: datetime):
        """
        listen room_topic event for puppet

        this is friendly for code typing
        """

    async def on_scan(self, status: ScanStatus, qr_code: Optional[str] = None,
                      data: Optional[str] = None):
        """
        listen scan event for puppet

        this is friendly for code typing
        """

    def get_output(self) -> dict:
        """if necessary , get the output of the plugin"""
        final_output = deepcopy(self.output)
        self.output = {}

        return final_output


PluginTree = Dict[str, Union[str, List[str]]]


class WechatyPluginManager:
    """manage the wechaty plugin, It will support some features."""
    def __init__(self, wechaty: Wechaty):
        self._plugins: Dict[str, WechatyPlugin] = OrderedDict()
        self._wechaty: Wechaty = wechaty
        self._plugin_status: Dict[str, PluginStatus] = {}
        # plugins can be a topological graph pattern, this feature is not
        # supported now.
        self._dependency_tree: PluginTree = defaultdict()

    # pylint: disable=R1711
    @staticmethod
    def _load_plugin_from_local_file(plugin_path: str) -> Optional[WechatyPlugin]:
        """load plugin from local file"""
        log.info('load plugin from local file <%s>', plugin_path)
        return None

    # pylint: disable=R1711
    @staticmethod
    def _load_plugin_from_github_url(github_url: str
                                     ) -> Optional[WechatyPlugin]:
        """load plugin from github url, but, this is dangerous"""
        log.info('load plugin from github url <%s>', github_url)
        return None

    def add_plugin(self, plugin: Union[str, WechatyPlugin]):
        """add plugin to the manager, if the plugin name exist, it will not to
        be installed"""
        if isinstance(plugin, str):
            regex = re.compile(
                r'^(?:http|ftp)s?://'  # http:// or https://
                r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}'
                r'\.?|[A-Z0-9-]{2,}\.?)|'
                r'localhost|'  # localhost...
                r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
                r'(?::\d+)?'  # optional port
                r'(?:/?|[/?]\S+)$', re.IGNORECASE)
            if regex.match(plugin) is None:
                # load plugin from local file
                plugin_instance = self._load_plugin_from_local_file(plugin)
            else:
                plugin_instance = self._load_plugin_from_github_url(plugin)
            if plugin_instance is None:
                raise Exception('can"t load plugin %s' % plugin)
        else:
            if plugin.name in self._plugins:
                log.warning('plugin : %s has exist', plugin.name)
                return
            plugin_instance = plugin

        self._plugins[plugin_instance.name] = plugin_instance
        # default wechaty plugin status is Running
        self._plugin_status[plugin_instance.name] = PluginStatus.Running

    def remove_plugin(self, name: str):
        """remove plugin"""
        if name not in self._plugins:
            raise IndexError(f'plugin {name} not exist')
        self._plugins.pop(name)
        self._plugin_status.pop(name)

    def _check_plugins(self, name: str):
        """
        check the plugins whether
        """
        if name not in self._plugins and name not in self._plugin_status:
            raise Exception('plugins <%s> not exist' % name)

    def stop_plugin(self, name: str):
        """stop the plugin"""
        log.info('stopping the plugin <%s>', name)
        self._check_plugins(name)

        if self._plugin_status[name] == PluginStatus.Stopped:
            log.warning('plugins <%s> is stopped', name)
        self._plugin_status[name] = PluginStatus.Stopped

    def start_plugin(self, name: str):
        """starting the plugin"""
        log.info('starting the plugin <%s>', name)
        self._check_plugins(name)
        self._plugin_status[name] = PluginStatus.Running

    def plugin_status(self, name: str) -> PluginStatus:
        """get the plugin status"""
        self._check_plugins(name)
        return self._plugin_status[name]

    async def init_plugins(self):
        """
        set wechaty to plugins
        """
        log.info('init the plugins ...')
        for name, plugin in self._plugins.items():
            log.info('init %s-plugin ...', name)
            await plugin.init_plugin(self._wechaty)

    async def emit_events(self, event_name: str, *args, **kwargs):
        """

        during the try-stage, only support message_events

        event_name: get event
        event_payload:
        """
        if event_name == 'message':
            # this is not type linting, it's needed to be imported at top-level
            # , So, it must occurs cyclic-import problems, to resolve this, a
            # simple way is that we import package at local-level.

            # pylint: disable=import-outside-toplevel
            from wechaty import Message
            if len(args) == 1 and isinstance(args[0], Message):
                msg: Message = args[0]
            elif 'msg' in kwargs and isinstance(kwargs['msg'], Message):
                msg = kwargs['msg']
            else:
                raise ValueError('can"t find the message params')

            # this will make the plugins running sequential, _plugins
            # is a sort dict
            for name, plugin in self._plugins.items():
                log.info('emit %s-plugin ...', name)
                if self.plugin_status(name) == PluginStatus.Running:
                    await plugin.on_message(msg)



"""check the weather"""
from typing import Union

import requests

from wechaty import Message, Contact, Room
from wechaty.plugin import WechatyPlugin


class WeatherPlugin(WechatyPlugin):
    """weather plugin for bot"""
    @property
    def name(self) -> str:
        """get the name of the plugin"""
        return 'weather'

    async def on_message(self, msg: Message):
        """listen message event"""
        from_contact = msg.talker()
        text = msg.text()
        room = msg.room()
        if text == '今天天气如何':
            conversation: Union[
                Room, Contact] = from_contact if room is None else room
            await conversation.ready()
            response = requests.get('https://tianqiapi.com/api?version=v61&'
                                    'appid=32896971&appsecret=5bR8Gs9x')
            result = response.json()
            result_msg = f'今天{result["wea"]} 最低温度{result["tem2"]}度 ' \
                         f'最高温度{result["tem1"]}度'
            await conversation.say(result_msg)



"""basic ding-dong bot for the wechaty plugin"""
from typing import Union

from wechaty import Message, Contact, Room, FileBox
from wechaty.plugin import WechatyPlugin


class DingDongPlugin(WechatyPlugin):
    """basic ding-dong plugin"""
    @property
    def name(self):
        """name of the plugin"""
        return 'ding-dong'

    async def on_message(self, msg: Message):
        """listen message event"""
        from_contact = msg.talker()
        text = msg.text()
        room = msg.room()
        if text == '#ding':
            conversation: Union[
                Room, Contact] = from_contact if room is None else room
            await conversation.ready()
            await conversation.say('dong')
            file_box = FileBox.from_url(
                'https://ss3.bdstatic.com/70cFv8Sh_Q1YnxGkpoWK1HF6hhy/it/'
                'u=1116676390,2305043183&fm=26&gp=0.jpg',
                name='ding-dong.jpg')
            await conversation.say(file_box)



"""daily plugin"""
from datetime import datetime

from apscheduler.schedulers.asyncio import AsyncIOScheduler     # type: ignore

from wechaty import Wechaty
from wechaty.plugin import WechatyPlugin


class DailyPlugin(WechatyPlugin):
    """
    say something everyday, like `Daily Words`
    """
    @property
    def name(self) -> str:
        """get the name of the plugin"""
        return 'dayily'

    async def tick(self):
        """time tick for the plugin scheduler"""
        room = self.bot.Room.load('19961884194@chatroom')
        await room.ready()
        await room.say(f'i love you -> {datetime.now()}')

    async def init_plugin(self, wechaty: Wechaty):
        """init plugin"""
        await super().init_plugin(wechaty)
        scheduler = AsyncIOScheduler()
        scheduler.add_job(self.tick, 'interval', seconds=10)
        scheduler.start()



"""import all plugins"""
from wechaty_plugin_contrib.daily_plugin import DailyPlugin
from wechaty_plugin_contrib.ding_dong_plugin import DingDongPlugin
from wechaty_plugin_contrib.weather_plugin import WeatherPlugin


__all__ = [
    'DailyPlugin',
    'DingDongPlugin',
    'WeatherPlugin'
]



"""
Python Wechaty - https://github.com/wechaty/python-wechaty

Authors:    Huan LI (李卓桓) <https://github.com/huan>
            Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2020-now @ Copyright Wechaty

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations

from wechaty_puppet import MiniProgramPayload, get_logger   # type: ignore


log = get_logger('MiniProgram')


class MiniProgram:
    """
    mini_program object which handle the url_link content
    """
    def __init__(self, payload: MiniProgramPayload):
        """
        initialization for mini_program
        :param payload:
        """
        log.info('MiniProgram created')
        self.payload: MiniProgramPayload = payload

    @classmethod
    async def create(cls) -> MiniProgram:
        """
        static create MiniProgram method
        :return:
        """
        log.info('MiniProgram created')
        # TODO -> create default mini_program payload
        payload = MiniProgramPayload()
        return MiniProgram(payload)

    @property
    def app_id(self) -> str:
        """
        get mini_program app_id
        :return:
        """
        if self.payload.app_id is None:
            return ''
        return self.payload.app_id

    @property
    def title(self) -> str:
        """
        get mini_program title
        :return:
        """
        if self.payload.title is None:
            return ''
        return self.payload.title

    @property
    def page_path(self) -> str:
        """
        get mini_program page_path
        :return:
        """
        if self.payload.page_path is None:
            return ''
        return self.payload.page_path

    @property
    def user_name(self) -> str:
        """
        get mini_program user_name
        :return:
        """
        if self.payload.user_name is None:
            return ''
        return self.payload.user_name

    @property
    def description(self) -> str:
        """
        get mini_program description
        :return:
        """
        if self.payload.description is None:
            return ''
        return self.payload.description

    @property
    def thumb_url(self) -> str:
        """
        get mini_program thumb_url
        :return:
        """
        if self.payload.thumb_url is None:
            return ''
        return self.payload.thumb_url

    @property
    def thumb_key(self) -> str:
        """
        get mini_program thumb_key
        :return:
        """
        if self.payload.thumb_key is None:
            return ''
        return self.payload.thumb_key
